@@ -1,6 +1,6 @@
-/*
+/* 
  * dhcpcd - DHCP client daemon
- * Copyright (c) 2006-2013 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2012 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -33,30 +33,26 @@
 #include <netinet/ip6.h>
 #include <netinet/icmp6.h>
 
-#ifdef __linux__
-#  define _LINUX_IN6_H
-#  include <linux/ipv6.h>
-#endif
-
 #include <errno.h>
 #include <stddef.h>
 #include <stdlib.h>
 #include <string.h>
 #include <syslog.h>
-#include <unistd.h>
 
-#define ELOOP_QUEUE 2
+#ifdef __linux__
+#  define _LINUX_IN6_H
+#  include <linux/ipv6.h>
+#endif
+
+#define ELOOP_QUEUE 1
+#include "bind.h"
 #include "common.h"
+#include "configure.h"
 #include "dhcpcd.h"
-#include "dhcp6.h"
 #include "eloop.h"
 #include "ipv6.h"
 #include "ipv6ns.h"
 #include "ipv6rs.h"
-#include "script.h"
-
-/* Debugging Router Solicitations is a lot of spam, so disable it */
-//#define DEBUG_RS
 
 #define RTR_SOLICITATION_INTERVAL       4 /* seconds */
 #define MAX_RTR_SOLICITATIONS           3 /* times */
@@ -69,7 +65,7 @@ struct nd_opt_rdnss {           /* RDNSS option RFC 6106 */
 	uint16_t	nd_opt_rdnss_reserved;
 	uint32_t	nd_opt_rdnss_lifetime;
         /* followed by list of IP prefixes */
-} __packed;
+} _packed;
 #endif
 
 #ifndef ND_OPT_DNSSL
@@ -80,7 +76,7 @@ struct nd_opt_dnssl {		/* DNSSL option RFC 6106 */
 	uint16_t	nd_opt_dnssl_reserved;
 	uint32_t	nd_opt_dnssl_lifetime;
 	/* followed by list of DNS servers */
-} __packed;
+} _packed;
 #endif
 
 /* Minimal IPv6 MTU */
@@ -105,7 +101,7 @@ struct nd_opt_dnssl {		/* DNSSL option RFC 6106 */
 
 struct rahead ipv6_routers = TAILQ_HEAD_INITIALIZER(ipv6_routers);
 
-static int sock = -1;
+static int sock;
 static struct sockaddr_in6 allrouters, from;
 static struct msghdr sndhdr;
 static struct iovec sndiov[2];
@@ -126,39 +122,38 @@ ipv6rs_cleanup(void)
 }
 #endif
 
-static int
+int
 ipv6rs_open(void)
 {
 	int on;
 	int len;
 	struct icmp6_filter filt;
 
-	sock = socket(AF_INET6, SOCK_RAW, IPPROTO_ICMPV6);
-	if (sock == -1)
-		return -1;
-
 	memset(&allrouters, 0, sizeof(allrouters));
 	allrouters.sin6_family = AF_INET6;
 #ifdef SIN6_LEN
 	allrouters.sin6_len = sizeof(allrouters);
 #endif
 	if (inet_pton(AF_INET6, ALLROUTERS, &allrouters.sin6_addr.s6_addr) != 1)
-		goto eexit;
+		return -1;
+	sock = socket(AF_INET6, SOCK_RAW, IPPROTO_ICMPV6);
+	if (sock == -1)
+		return -1;
 	on = 1;
 	if (setsockopt(sock, IPPROTO_IPV6, IPV6_RECVPKTINFO,
 		&on, sizeof(on)) == -1)
-		goto eexit;
+		return -1;
 
 	on = 1;
 	if (setsockopt(sock, IPPROTO_IPV6, IPV6_RECVHOPLIMIT,
 		&on, sizeof(on)) == -1)
-		goto eexit;
+		return -1;
 
 	ICMP6_FILTER_SETBLOCKALL(&filt);
 	ICMP6_FILTER_SETPASS(ND_ROUTER_ADVERT, &filt);
 	if (setsockopt(sock, IPPROTO_ICMPV6, ICMP6_FILTER,
 		&filt, sizeof(filt)) == -1)
-		goto eexit;
+		return -1;
 
 	set_cloexec(sock);
 #if DEBUG_MEMORY
@@ -166,17 +161,17 @@ ipv6rs_open(void)
 #endif
 
 	len = CMSG_SPACE(sizeof(struct in6_pktinfo)) + CMSG_SPACE(sizeof(int));
-	sndbuf = calloc(1, len);
+	sndbuf = xzalloc(len);
 	if (sndbuf == NULL)
-		goto eexit;
+		return -1;
 	sndhdr.msg_namelen = sizeof(struct sockaddr_in6);
 	sndhdr.msg_iov = sndiov;
 	sndhdr.msg_iovlen = 1;
 	sndhdr.msg_control = sndbuf;
 	sndhdr.msg_controllen = len;
-	rcvbuf = calloc(1, len);
+	rcvbuf = xzalloc(len);
 	if (rcvbuf == NULL)
-		goto eexit;
+		return -1;
 	rcvhdr.msg_name = &from;
 	rcvhdr.msg_namelen = sizeof(from);
 	rcvhdr.msg_iov = rcviov;
@@ -186,67 +181,47 @@ ipv6rs_open(void)
 	rcviov[0].iov_base = ansbuf;
 	rcviov[0].iov_len = sizeof(ansbuf);
 	return sock;
-
-eexit:
-	close(sock);
-	sock = -1;
-	free(sndbuf);
-	sndbuf = NULL;
-	free(rcvbuf);
-	rcvbuf = NULL;
-	return -1;
 }
 
 static int
 ipv6rs_makeprobe(struct interface *ifp)
 {
-	struct rs_state *state;
 	struct nd_router_solicit *rs;
 	struct nd_opt_hdr *nd;
 
-	state = RS_STATE(ifp);
-	free(state->rs);
-	state->rslen = sizeof(*rs) + ROUNDUP8(ifp->hwlen + 2);
-	state->rs = calloc(1, state->rslen);
-	if (state->rs == NULL)
+	free(ifp->rs);
+	ifp->rslen = sizeof(*rs) + ROUNDUP8(ifp->hwlen + 2);
+	ifp->rs = xzalloc(ifp->rslen);
+	if (ifp->rs == NULL)
 		return -1;
-	rs = (struct nd_router_solicit *)(void *)state->rs;
+	rs = (struct nd_router_solicit *)(void *)ifp->rs;
 	rs->nd_rs_type = ND_ROUTER_SOLICIT;
 	rs->nd_rs_code = 0;
 	rs->nd_rs_cksum = 0;
 	rs->nd_rs_reserved = 0;
-	nd = (struct nd_opt_hdr *)(state->rs + sizeof(*rs));
+	nd = (struct nd_opt_hdr *)(ifp->rs + sizeof(*rs));
 	nd->nd_opt_type = ND_OPT_SOURCE_LINKADDR;
 	nd->nd_opt_len = (ROUNDUP8(ifp->hwlen + 2)) >> 3;
 	memcpy(nd + 1, ifp->hwaddr, ifp->hwlen);
 	return 0;
 }
-
+	
 static void
 ipv6rs_sendprobe(void *arg)
 {
 	struct interface *ifp = arg;
-	struct rs_state *state;
 	struct sockaddr_in6 dst;
 	struct cmsghdr *cm;
 	struct in6_pktinfo pi;
 	int hoplimit = HOPLIMIT;
 
-	if (ipv6_linklocal(ifp) == NULL) {
-		syslog(LOG_DEBUG,
-		    "%s: delaying Router Soliciation for LL address",
-		    ifp->name);
-		ipv6_addlinklocalcallback(ifp, ipv6rs_sendprobe, ifp);
-		return;
-	}
-
 	dst = allrouters;
-	dst.sin6_scope_id = ifp->index;
+	//dst.sin6_scope_id = ifp->linkid;
 
-	state = RS_STATE(ifp);
+	ipv6rs_makeprobe(ifp);
 	sndhdr.msg_name = (caddr_t)&dst;
-	sndhdr.msg_iov[0].iov_base = state->rs;
-	sndhdr.msg_iov[0].iov_len = state->rslen;
+	sndhdr.msg_iov[0].iov_base = ifp->rs;
+	sndhdr.msg_iov[0].iov_len = ifp->rslen;
 
 	/* Set the outbound interface */
 	cm = CMSG_FIRSTHDR(&sndhdr);
@@ -264,19 +239,15 @@ ipv6rs_sendprobe(void *arg)
 	cm->cmsg_len = CMSG_LEN(sizeof(hoplimit));
 	memcpy(CMSG_DATA(cm), &hoplimit, sizeof(hoplimit));
 
-	syslog(LOG_DEBUG, "%s: sending Router Solicitation", ifp->name);
-	if (sendmsg(sock, &sndhdr, 0) == -1) {
-		syslog(LOG_ERR, "%s: %s: sendmsg: %m", ifp->name, __func__);
-		ipv6rs_drop(ifp);
-		ifp->options->options &= ~(DHCPCD_IPV6 | DHCPCD_IPV6RS);
-		return;
-	}
+	syslog(LOG_INFO, "%s: sending IPv6 Router Solicitation", ifp->name);
+	if (sendmsg(sock, &sndhdr, 0) == -1)
+		syslog(LOG_ERR, "%s: sendmsg: %m", ifp->name);
 
-	if (state->rsprobes++ < MAX_RTR_SOLICITATIONS)
-		eloop_timeout_add_sec(RTR_SOLICITATION_INTERVAL,
+	if (ifp->rsprobes++ < MAX_RTR_SOLICITATIONS)
+		add_timeout_sec(RTR_SOLICITATION_INTERVAL,
 		    ipv6rs_sendprobe, ifp);
 	else
-		syslog(LOG_WARNING, "%s: no IPv6 Routers available", ifp->name);
+		syslog(LOG_INFO, "%s: no IPv6 Routers available", ifp->name);
 }
 
 static void
@@ -291,29 +262,30 @@ ipv6rs_free_opts(struct ra *rap)
 	}
 }
 
-int
-ipv6rs_addrexists(const struct ipv6_addr *a)
+static void
+ipv6rs_freedrop_addrs(struct ra *rap, int drop)
 {
-	struct ra *rap;
 	struct ipv6_addr *ap;
 
-	TAILQ_FOREACH(rap, &ipv6_routers, next) {
-		TAILQ_FOREACH(ap, &rap->addrs, next) {
-			if (memcmp(&ap->addr, &a->addr, sizeof(a->addr)) == 0)
-				return 1;
+	while ((ap = TAILQ_FIRST(&rap->addrs))) {
+		TAILQ_REMOVE(&rap->addrs, ap, next);
+		if (drop && (options & DHCPCD_IPV6RA_OWN)) {
+			syslog(LOG_INFO, "%s: deleting address %s",
+			    rap->iface->name, ap->saddr);
+			if (del_address6(rap->iface, ap) == -1)
+				syslog(LOG_ERR, "del_address6 %m");
 		}
+		free(ap);
 	}
-	return 0;
 }
 
 void ipv6rs_freedrop_ra(struct ra *rap, int drop)
 {
 
-	eloop_timeout_delete(NULL, rap->iface);
-	eloop_timeout_delete(NULL, rap);
-	if (!drop)
-		TAILQ_REMOVE(&ipv6_routers, rap, next);
-	ipv6_freedrop_addrs(&rap->addrs, drop, NULL);
+	delete_timeout(NULL, rap->iface);
+	delete_timeout(NULL, rap);
+	TAILQ_REMOVE(&ipv6_routers, rap, next);
+	ipv6rs_freedrop_addrs(rap, drop);
 	ipv6rs_free_opts(rap);
 	free(rap->data);
 	free(rap->ns);
@@ -323,16 +295,11 @@ void ipv6rs_freedrop_ra(struct ra *rap, int drop)
 ssize_t
 ipv6rs_free(struct interface *ifp)
 {
-	struct rs_state *state;
 	struct ra *rap, *ran;
 	ssize_t n;
 
-	state = RS_STATE(ifp);
-	if (state) {
-		free(state->rs);
-		free(state);
-		ifp->if_data[IF_DATA_IPV6RS] = NULL;
-	}
+	free(ifp->rs);
+	ifp->rs = NULL;
 	n = 0;
 	TAILQ_FOREACH_SAFE(rap, &ipv6_routers, next, ran) {
 		if (rap->iface == ifp) {
@@ -346,7 +313,6 @@ ipv6rs_free(struct interface *ifp)
 static int
 rtpref(struct ra *rap)
 {
-
 	switch (rap->flags & ND_RA_FLAG_RTPREF_MASK) {
 	case ND_RA_FLAG_RTPREF_HIGH:
 		return (RTPREF_HIGH);
@@ -376,117 +342,14 @@ add_router(struct ra *router)
 			return;
 		}
 	}
-	TAILQ_INSERT_TAIL(&ipv6_routers, router, next);
-}
-
-static void
-ipv6rs_scriptrun(struct ra *rap)
-{
-	int hasdns;
-	struct ipv6_addr *ap;
-	const struct ra_opt *rao;
-
-	/* If all addresses have completed DAD run the script */
-	TAILQ_FOREACH(ap, &rap->addrs, next) {
-		if ((ap->flags & (IPV6_AF_ONLINK | IPV6_AF_AUTOCONF)) ==
-		    (IPV6_AF_ONLINK | IPV6_AF_AUTOCONF))
-		{
-			if (!(ap->flags & IPV6_AF_DADCOMPLETED) &&
-			    ipv6_findaddr(ap->iface, &ap->addr))
-				ap->flags |= IPV6_AF_DADCOMPLETED;
-			if ((ap->flags & IPV6_AF_DADCOMPLETED) == 0) {
-				syslog(LOG_DEBUG,
-				    "%s: waiting for Router Advertisement"
-				    " DAD to complete",
-				    rap->iface->name);
-				return;
-			}
-		}
-	}
-
-	/* If we don't require RDNSS then set hasdns = 1 so we fork */
-	if (!(rap->iface->options->options & DHCPCD_IPV6RA_REQRDNSS))
-		hasdns = 1;
-	else {
-		hasdns = 0;
-		TAILQ_FOREACH(rao, &rap->options, next) {
-			if (rao->type == ND_OPT_RDNSS &&
-			    rao->option &&
-			    timerisset(&rao->expire))
-			{
-				hasdns = 1;
-				break;
-			}
-		}
-	}
-
-	script_runreason(rap->iface, "ROUTERADVERT");
-	if (hasdns)
-		daemonise();
-#if 0
-	else if (options & DHCPCD_DAEMONISE &&
-	    !(options & DHCPCD_DAEMONISED) && new_data)
-		syslog(LOG_WARNING,
-		    "%s: did not fork due to an absent"
-		    " RDNSS option in the RA",
-		    ifp->name);
-}
-#endif
-}
-
-static void
-ipv6rs_dadcallback(void *arg)
-{
-	struct ipv6_addr *ap = arg, *rapap;
-	struct interface *ifp;
-	struct ra *rap;
-	int wascompleted, found;
-
-	wascompleted = (ap->flags & IPV6_AF_DADCOMPLETED);
-	ipv6ns_cancelprobeaddr(ap);
-	ap->flags |= IPV6_AF_DADCOMPLETED;
-	if (ap->flags & IPV6_AF_DUPLICATED)
-		/* No idea what how to try and make another address :( */
-		syslog(LOG_WARNING, "%s: DAD detected %s",
-		    ap->iface->name, ap->saddr);
-#ifdef IPV6_SEND_DAD
-	else
-		ipv6_addaddr(ap);
-#endif
-
-	if (!wascompleted) {
-		ifp = ap->iface;
-
-		TAILQ_FOREACH(rap, &ipv6_routers, next) {
-			if (rap->iface != ifp)
-				continue;
-			wascompleted = 1;
-			TAILQ_FOREACH(rapap, &rap->addrs, next) {
-				if (rapap->flags & IPV6_AF_AUTOCONF &&
-				    (rapap->flags & IPV6_AF_DADCOMPLETED) == 0)
-				{
-					wascompleted = 0;
-					break;
-				}
-				if (rapap == ap)
-					found = 1;
-			}
-
-			if (wascompleted && found && rap->lifetime) {
-				syslog(LOG_DEBUG,
-				    "%s: Router Advertisement DAD completed",
-				    rap->iface->name);
-				ipv6rs_scriptrun(rap);
-			}
-		}
-	}
+	TAILQ_INSERT_HEAD(&ipv6_routers, router, next);
 }
 
 /* ARGSUSED */
-static void
-ipv6rs_handledata(__unused void *arg)
+void
+ipv6rs_handledata(_unused void *arg)
 {
-	ssize_t len, l, m, n, olen;
+	ssize_t len, l, n, olen;
 	struct cmsghdr *cm;
 	int hoplimit;
 	struct in6_pktinfo pkt;
@@ -507,9 +370,9 @@ ipv6rs_handledata(__unused void *arg)
 	struct nd_opt_hdr *ndo;
 	struct ra_opt *rao;
 	struct ipv6_addr *ap;
-	char *opt, *tmp;
+	char *opt;
 	struct timeval expire;
-	uint8_t new_rap, new_data;
+	int has_dns, new_rap;
 
 	len = recvmsg(sock, &rcvhdr, 0);
 	if (len == -1) {
@@ -562,72 +425,35 @@ ipv6rs_handledata(__unused void *arg)
 		return;
 	}
 
-	TAILQ_FOREACH(ifp, ifaces, next) {
+	for (ifp = ifaces; ifp; ifp = ifp->next)
 		if (ifp->index == (unsigned int)pkt.ipi6_ifindex)
 			break;
-	}
 	if (ifp == NULL) {
-#ifdef DEBUG_RS
-		syslog(LOG_DEBUG, "RA for unexpected interface from %s", sfrom);
-#endif
+		syslog(LOG_ERR, "RA for unexpected interface from %s", sfrom);
 		return;
 	}
-	if (!(ifp->options->options & DHCPCD_IPV6RS)) {
-#ifdef DEBUG_RS
-		syslog(LOG_DEBUG, "%s: unexpected RA from %s",
-		    ifp->name, sfrom);
-#endif
-		return;
-	}
-
-	/* We could recieve a RA before we sent a RS*/
-	if (ipv6_linklocal(ifp) == NULL) {
-#ifdef DEBUG_RS
-		syslog(LOG_DEBUG, "%s: received RA from %s (no link-local)",
-		    ifp->name, sfrom);
-#endif
-		return;
-	}
-
 	TAILQ_FOREACH(rap, &ipv6_routers, next) {
-		if (ifp == rap->iface &&
-		    memcmp(rap->from.s6_addr, from.sin6_addr.s6_addr,
+		if (memcmp(rap->from.s6_addr, from.sin6_addr.s6_addr,
 		    sizeof(rap->from.s6_addr)) == 0)
 			break;
 	}
 
-	nd_ra = (struct nd_router_advert *)icp;
-	/* Don't bother doing anything if we don't know about a router
-	 * expiring */
-	if ((rap == NULL || rap->lifetime == 0)
-	    && nd_ra->nd_ra_router_lifetime == 0)
-		return;
-
 	/* We don't want to spam the log with the fact we got an RA every
 	 * 30 seconds or so, so only spam the log if it's different. */
-	if (rap == NULL || (rap->data_len != len ||
+	if (options & DHCPCD_DEBUG || rap == NULL ||
+	    (rap->expired || rap->data_len != len ||
 	     memcmp(rap->data, (unsigned char *)icp, rap->data_len) != 0))
 	{
 		if (rap) {
 			free(rap->data);
 			rap->data_len = 0;
-			free(rap->ns);
-			rap->ns = NULL;
-			rap->nslen = 0;
 		}
-		new_data = 1;
-	} else
-		new_data = 0;
-	if (new_data || ifp->options->options & DHCPCD_DEBUG)
 		syslog(LOG_INFO, "%s: Router Advertisement from %s",
 		    ifp->name, sfrom);
+	}
 
 	if (rap == NULL) {
-		rap = calloc(1, sizeof(*rap));
-		if (rap == NULL) {
-			syslog(LOG_ERR, "%s: %m", __func__);
-			return;
-		}
+		rap = xzalloc(sizeof(*rap));
 		rap->iface = ifp;
 		memcpy(rap->from.s6_addr, from.sin6_addr.s6_addr,
 		    sizeof(rap->from.s6_addr));
@@ -638,37 +464,22 @@ ipv6rs_handledata(__unused void *arg)
 	} else
 		new_rap = 0;
 	if (rap->data_len == 0) {
-		rap->data = malloc(len);
-		if (rap->data == NULL) {
-			syslog(LOG_ERR, "%s: %m", __func__);
-			if (new_rap)
-				free(rap);
-			return;
-		}
+		rap->data = xmalloc(len);
 		memcpy(rap->data, icp, len);
 		rap->data_len = len;
 	}
 
 	get_monotonic(&rap->received);
+	nd_ra = (struct nd_router_advert *)icp;
 	rap->flags = nd_ra->nd_ra_flags_reserved;
-	if (new_rap == 0 && rap->lifetime == 0)
-		syslog(LOG_WARNING, "%s: %s router available",
-		   ifp->name, rap->sfrom);
 	rap->lifetime = ntohs(nd_ra->nd_ra_router_lifetime);
-	if (nd_ra->nd_ra_reachable) {
-		rap->reachable = ntohl(nd_ra->nd_ra_reachable);
-		if (rap->reachable > MAX_REACHABLE_TIME)
-			rap->reachable = 0;
-	}
-	if (nd_ra->nd_ra_retransmit)
-		rap->retrans = ntohl(nd_ra->nd_ra_retransmit);
-	if (rap->lifetime)
-		rap->expired = 0;
+	rap->expired = 0;
 
 	len -= sizeof(struct nd_router_advert);
 	p = ((uint8_t *)icp) + sizeof(struct nd_router_advert);
 	olen = 0;
 	lifetime = ~0U;
+	has_dns = 0;
 	for (olen = 0; len > 0; p += olen, len -= olen) {
 		if ((size_t)len < sizeof(struct nd_opt_hdr)) {
 			syslog(LOG_ERR, "%s: Short option", ifp->name);
@@ -708,13 +519,6 @@ ipv6rs_handledata(__unused void *arg)
 				    "%s: invalid prefix in RA", ifp->name);
 				break;
 			}
-			if (ntohl(pi->nd_opt_pi_preferred_time) >
-			    ntohl(pi->nd_opt_pi_valid_time))
-			{
-				syslog(LOG_ERR,
-				    "%s: pltime > vltime", ifp->name);
-				break;
-			}
 			TAILQ_FOREACH(ap, &rap->addrs, next)
 				if (ap->prefix_len ==pi->nd_opt_pi_prefix_len &&
 				    memcmp(ap->prefix.s6_addr,
@@ -722,66 +526,33 @@ ipv6rs_handledata(__unused void *arg)
 				    sizeof(ap->prefix.s6_addr)) == 0)
 					break;
 			if (ap == NULL) {
-				if (!(pi->nd_opt_pi_flags_reserved &
-				    ND_OPT_PI_FLAG_AUTO) &&
-				    !(pi->nd_opt_pi_flags_reserved &
-				    ND_OPT_PI_FLAG_ONLINK))
-					break;
-				ap = calloc(1, sizeof(*ap));
-				if (ap == NULL) {
-					syslog(LOG_ERR, "%s: %m", __func__);
-					break;
-				}
-				ap->iface = rap->iface;
-				ap->flags = IPV6_AF_NEW;
+				ap = xmalloc(sizeof(*ap));
+				ap->new = 1;
 				ap->prefix_len = pi->nd_opt_pi_prefix_len;
 				memcpy(ap->prefix.s6_addr,
 				   pi->nd_opt_pi_prefix.s6_addr,
 				   sizeof(ap->prefix.s6_addr));
-				if (pi->nd_opt_pi_flags_reserved &
-				    ND_OPT_PI_FLAG_AUTO)
-				{
-					ap->flags |= IPV6_AF_AUTOCONF;
-					ipv6_makeaddr(&ap->addr, ifp,
-					    &ap->prefix,
-					    pi->nd_opt_pi_prefix_len);
-					cbp = inet_ntop(AF_INET6,
-					    ap->addr.s6_addr,
-					    ntopbuf, INET6_ADDRSTRLEN);
-					if (cbp)
-						snprintf(ap->saddr,
-						    sizeof(ap->saddr),
-						    "%s/%d",
-						    cbp, ap->prefix_len);
-					else
-						ap->saddr[0] = '\0';
-				} else {
-					memset(&ap->addr, 0, sizeof(ap->addr));
+				ipv6_makeaddr(&ap->addr, ifp->name,
+				    &ap->prefix, pi->nd_opt_pi_prefix_len);
+				cbp = inet_ntop(AF_INET6, ap->addr.s6_addr,
+				    ntopbuf, INET6_ADDRSTRLEN);
+				if (cbp)
+					memcpy(ap->saddr, cbp,
+					    sizeof(ap->saddr));
+				else
 					ap->saddr[0] = '\0';
-				}
-				ap->dadcallback = ipv6rs_dadcallback;
 				TAILQ_INSERT_TAIL(&rap->addrs, ap, next);
-			}
-			if (pi->nd_opt_pi_flags_reserved &
-			    ND_OPT_PI_FLAG_ONLINK)
-				ap->flags |= IPV6_AF_ONLINK;
+			} else if (ap->prefix_vltime !=
+			    ntohl(pi->nd_opt_pi_valid_time) ||
+			    ap->prefix_pltime !=
+			    ntohl(pi->nd_opt_pi_preferred_time))
+				ap->new = 1;
+			else
+				ap->new = 0;
 			ap->prefix_vltime =
 			    ntohl(pi->nd_opt_pi_valid_time);
 			ap->prefix_pltime =
 			    ntohl(pi->nd_opt_pi_preferred_time);
-			ap->nsprobes = 0;
-			if (opt) {
-				l = strlen(opt);
-				tmp = realloc(opt,
-					l + strlen(ap->saddr) + 2);
-				if (tmp) {
-					opt = tmp;
-					opt[l] = ' ';
-					strcpy(opt + l + 1, ap->saddr);
-				}
-			} else
-				opt = strdup(ap->saddr);
-			lifetime = ap->prefix_vltime;
 			break;
 
 		case ND_OPT_MTU:
@@ -792,9 +563,10 @@ ipv6rs_handledata(__unused void *arg)
 				    ifp->name, mtuv);
 				break;
 			}
-			rap->mtu = mtuv;
+			if (rap->mtu == 0 || mtuv < rap->mtu)
+				rap->mtu = mtuv;
 			snprintf(buf, sizeof(buf), "%d", mtuv);
-			opt = strdup(buf);
+			opt = xstrdup(buf);
 			break;
 
 		case ND_OPT_RDNSS:
@@ -805,37 +577,30 @@ ipv6rs_handledata(__unused void *arg)
 			    nd_opt_rdnss_lifetime);
 			op += sizeof(rdnss->nd_opt_rdnss_lifetime);
 			l = 0;
-			for (n = ndo->nd_opt_len - 1; n > 1; n -= 2,
-			    op += sizeof(addr.s6_addr))
-			{
-				m = ipv6_printaddr(NULL, 0, op, ifp->name);
-				if (m != -1)
-					l += m + 1;
-			}
-			op = (uint8_t *)ndo;
-			op += offsetof(struct nd_opt_rdnss,
-			    nd_opt_rdnss_lifetime);
-			op += sizeof(rdnss->nd_opt_rdnss_lifetime);
-			tmp = opt = malloc(l);
-			if (opt) {
-				for (n = ndo->nd_opt_len - 1; n > 1; n -= 2,
-				    op += sizeof(addr.s6_addr))
-				{
-					m = ipv6_printaddr(tmp, l, op,
+			for (n = ndo->nd_opt_len - 1; n > 1; n -= 2) {
+				memcpy(&addr.s6_addr, op, sizeof(addr.s6_addr));
+				cbp = inet_ntop(AF_INET6, &addr,
+				    ntopbuf, INET6_ADDRSTRLEN);
+				if (cbp == NULL) {
+					syslog(LOG_ERR,
+					    "%s: invalid RDNSS address",
 					    ifp->name);
-					if (m != -1) {
-						l -= (m + 1);
-						tmp += m;
-						*tmp++ = ' ';
-					}
+				} else {
+					if (opt) {
+						l = strlen(opt);
+						opt = xrealloc(opt,
+							l + strlen(cbp) + 2);
+						opt[l] = ' ';
+						strcpy(opt + l + 1, cbp);
+					} else
+						opt = xstrdup(cbp);
+					if (lifetime > 0)
+						has_dns = 1;
 				}
-				if (tmp != opt)
-					(*--tmp) = '\0';
-				else
-					*opt = '\0';
+		        	op += sizeof(addr.s6_addr);
 			}
 			break;
-
+			
 		case ND_OPT_DNSSL:
 			dnssl = (struct nd_opt_dnssl *)p;
 			lifetime = ntohl(dnssl->nd_opt_dnssl_lifetime);
@@ -848,29 +613,14 @@ ipv6rs_handledata(__unused void *arg)
 				syslog(LOG_ERR, "%s: invalid DNSSL option",
 				    ifp->name);
 			} else {
-				tmp = malloc(l);
-				if (tmp) {
-					decode_rfc3397(tmp, l, n, op);
-					n = print_string(NULL, 0,
-					    l - 1, (const uint8_t *)tmp);
-					opt = malloc(n);
-					if (opt)
-						print_string(opt, n,
-						    l - 1,
-						    (const uint8_t *)tmp);
-					free(tmp);
-				}
+				opt = xmalloc(l);
+				decode_rfc3397(opt, l, n, op);
 			}
 			break;
-
-		default:
-			continue;
 		}
 
-		if (opt == NULL) {
-			syslog(LOG_ERR, "%s: %m", __func__);
+		if (opt == NULL)
 			continue;
-		}
 		TAILQ_FOREACH(rao, &rap->options, next) {
 			if (rao->type == ndo->nd_opt_type &&
 			    strcmp(rao->option, opt) == 0)
@@ -882,16 +632,11 @@ ipv6rs_handledata(__unused void *arg)
 				free(rao->option);
 				free(rao);
 			}
-			free(opt);
 			continue;
 		}
 
 		if (rao == NULL) {
-			rao = malloc(sizeof(*rao));
-			if (rao == NULL) {
-				syslog(LOG_ERR, "%s: %m", __func__);
-				continue;
-			}
+			rao = xmalloc(sizeof(*rao));
 			rao->type = ndo->nd_opt_type;
 			rao->option = opt;
 			TAILQ_INSERT_TAIL(&rap->options, rao, next);
@@ -908,50 +653,52 @@ ipv6rs_handledata(__unused void *arg)
 
 	if (new_rap)
 		add_router(rap);
-	if (options & DHCPCD_TEST) {
-		script_runreason(ifp, "TEST");
-		goto handle_flag;
+	if (options & DHCPCD_IPV6RA_OWN) {
+		TAILQ_FOREACH(ap, &rap->addrs, next) {
+			syslog(ap->new ? LOG_INFO : LOG_DEBUG,
+			    "%s: adding address %s",
+			    ifp->name, ap->saddr);
+			if (add_address6(ifp, ap) == -1)
+				syslog(LOG_ERR, "add_address6 %m");
+			else if (ipv6_remove_subnet(rap, ap) == -1)
+				syslog(LOG_ERR, "ipv6_remove_subnet %m");
+			else
+				syslog(ap->new ? LOG_INFO : LOG_DEBUG,
+				    "%s: vltime %d seconds, pltime %d seconds",
+				    ifp->name, ap->prefix_vltime,
+				    ap->prefix_pltime);
+		}
 	}
-	if (ifp->options->options & DHCPCD_IPV6RA_OWN)
-		ipv6ns_probeaddrs(&rap->addrs);
-	ipv6_buildroutes();
+	ipv6_build_routes();
+	run_script_reason(ifp, options & DHCPCD_TEST ? "TEST" : "ROUTERADVERT");
+	if (options & DHCPCD_TEST)
+		exit(EXIT_SUCCESS);
 
-	/* We will get run by the expire function */
-	if (rap->lifetime)
-		ipv6rs_scriptrun(rap);
+	/* If we don't require RDNSS then set has_dns = 1 so we fork */
+	if (!(ifp->state->options->options & DHCPCD_IPV6RA_REQRDNSS))
+		has_dns = 1;
 
-	eloop_timeout_delete(NULL, ifp);
-	eloop_timeout_delete(NULL, rap); /* reachable timer */
+	if (has_dns)
+		delete_q_timeout(0, handle_exit_timeout, NULL);
+	delete_timeout(NULL, ifp);
+	delete_timeout(NULL, rap); /* reachable timer */
+	ipv6rs_expire(ifp);
+	if (has_dns)
+		daemonise();
+	else if (options & DHCPCD_DAEMONISE && !(options & DHCPCD_DAEMONISED))
+		syslog(LOG_WARNING,
+		    "%s: did not fork due to an absent RDNSS option in the RA",
+		    ifp->name);
 
 	/* If we're owning the RA then we need to try and ensure the
 	 * router is actually reachable */
-	if (ifp->options->options & DHCPCD_IPV6RA_OWN ||
-	    ifp->options->options & DHCPCD_IPV6RA_OWN_DEFAULT)
+	if (options & DHCPCD_IPV6RA_OWN ||
+	    options & DHCPCD_IPV6RA_OWN_DEFAULT)
 	{
 		rap->nsprobes = 0;
-		if (rap->lifetime)
-			ipv6ns_proberouter(rap);
+		add_timeout_sec(REACHABLE_TIME, ipv6ns_unreachable, rap);
+		add_timeout_sec(DELAY_FIRST_PROBE_TIME, ipv6ns_sendprobe, rap);
 	}
-
-handle_flag:
-	if (rap->flags & ND_RA_FLAG_MANAGED) {
-		if (rap->lifetime && new_data &&
-		    dhcp6_start(ifp, DH6S_INIT) == -1)
-			syslog(LOG_ERR, "dhcp6_start: %s: %m", ifp->name);
-	} else if (rap->flags & ND_RA_FLAG_OTHER) {
-		if (rap->lifetime && new_data &&
-		    dhcp6_start(ifp, DH6S_INFORM) == -1)
-			syslog(LOG_ERR, "dhcp6_start: %s: %m", ifp->name);
-	} else {
-		if (rap->lifetime && new_data)
-			syslog(LOG_DEBUG, "%s: No DHCPv6 instruction in RA",
-			    ifp->name);
-		if (options & DHCPCD_TEST)
-			exit(EXIT_SUCCESS);
-	}
-
-	/* Expire should be called last as the rap object could be destroyed */
-	ipv6rs_expire(ifp);
 }
 
 int
@@ -969,16 +716,14 @@ ssize_t
 ipv6rs_env(char **env, const char *prefix, const struct interface *ifp)
 {
 	ssize_t l;
-	size_t len;
 	struct timeval now;
 	const struct ra *rap;
 	const struct ra_opt *rao;
 	int i;
 	char buffer[32];
 	const char *optn;
-	char **pref, **mtu, **rdnss, **dnssl, ***var, *new;
 
-	i = 0;
+	i = 1;
 	l = 0;
 	get_monotonic(&now);
 	TAILQ_FOREACH(rap, &ipv6_routers, next) {
@@ -988,137 +733,75 @@ ipv6rs_env(char **env, const char *prefix, const struct interface *ifp)
 		if (env) {
 			snprintf(buffer, sizeof(buffer),
 			    "ra%d_from", i);
-			if (setvar(&env, prefix, buffer, rap->sfrom) == -1)
-				return -1;
+			setvar(&env, prefix, buffer, rap->sfrom);
 		}
 		l++;
 
-		pref = mtu = rdnss = dnssl = NULL;
 		TAILQ_FOREACH(rao, &rap->options, next) {
 			if (rao->option == NULL)
 				continue;
-			var = NULL;
-			switch(rao->type) {
+			if (env == NULL) {
+				switch (rao->type) {
+				case ND_OPT_PREFIX_INFORMATION:
+					l += 4;
+					break;
+				default:
+					l++;
+				}
+				continue;
+			}
+			switch (rao->type) {
 			case ND_OPT_PREFIX_INFORMATION:
 				optn = "prefix";
-				var = &pref;
 				break;
 			case ND_OPT_MTU:
 				optn = "mtu";
-				var = &mtu;
 				break;
 			case ND_OPT_RDNSS:
 				optn = "rdnss";
-				var = &rdnss;
 				break;
 			case ND_OPT_DNSSL:
 				optn = "dnssl";
-				var = &dnssl;
 				break;
 			default:
 				continue;
 			}
-			if (*var == NULL) {
-				*var = env ? env : &new;
-				l++;
-			} else if (env) {
-				/* With single only options, last one takes
-				 * precedence */
-				if (rao->type == ND_OPT_MTU) {
-					new = strchr(**var, '=');
-					if (new == NULL) {
-						syslog(LOG_ERR, "new is null");
-						continue;
-					} else
-						new++;
-					len = (new - **var) +
-					    strlen(rao->option) + 1;
-					if (len > strlen(**var))
-						new = realloc(**var, len);
-					else
-						new = **var;
-					if (new) {
-						**var = new;
-						new = strchr(**var, '=');
-						if (new)
-							strcpy(new + 1,
-							    rao->option);
-						else
-							syslog(LOG_ERR,
-							    "new is null");
-					}
-					continue;
-				}
-				new = realloc(**var,
-				    strlen(**var) + 1 +
-				    strlen(rao->option) + 1);
-				if (new == NULL)
-					return -1;
-				**var = new;
-				new += strlen(new);
-				*new++ = ' ';
-				strcpy(new, rao->option);
-				continue;
-			}
-			if (env) {
+			snprintf(buffer, sizeof(buffer), "ra%d_%s", i, optn);
+			setvar(&env, prefix, buffer, rao->option);
+			l++;
+#if 0
+			switch (rao->type) {
+			case ND_OPT_PREFIX_INFORMATION:
 				snprintf(buffer, sizeof(buffer),
-				    "ra%d_%s", i, optn);
-				if (setvar(&env, prefix, buffer, rao->option)
-				    == -1)
-					return -1;
+				    "ra%d_prefix_len", i);
+				snprintf(buffer2, sizeof(buffer2),
+				    "%d", rap->prefix_len);
+				setvar(&env, prefix, buffer, buffer2);
+
+				snprintf(buffer, sizeof(buffer),
+				    "ra%d_prefix_vltime", i);
+				snprintf(buffer2, sizeof(buffer2),
+				    "%d", rap->prefix_vltime);
+				setvar(&env, prefix, buffer, buffer2);
+
+				snprintf(buffer, sizeof(buffer),
+				    "ra%d_prefix_pltime", i);
+				snprintf(buffer2, sizeof(buffer2),
+				    "%d", rap->prefix_pltime);
+				setvar(&env, prefix, buffer, buffer2);
+				l += 3;
+				break;
 			}
+#endif		
 		}
 	}
 
-	if (env) {
-		if (setvard(&env, prefix, "ra_count", i) == -1)
-			return -1;
-	}
+	if (env)
+		setvard(&env, prefix, "ra_count", i - 1);
 	l++;
 	return l;
 }
 
-const struct ipv6_addr *
-ipv6rs_findprefix(const struct ipv6_addr *a)
-{
-	const struct ra *rap;
-	const struct ipv6_addr *ap;
-	const struct in6_addr *p1, *p2;
-	int bytelen, bitlen;
-
-	p1 = &a->addr;
-	TAILQ_FOREACH(rap, &ipv6_routers, next) {
-		TAILQ_FOREACH(ap, &rap->addrs, next) {
-			p2 = & ap->prefix;
-			bytelen = ap->prefix_len / NBBY;
-			bitlen = ap->prefix_len % NBBY;
-			if (memcmp(&p1->s6_addr, &p2->s6_addr, bytelen))
-				continue;
-			if (bitlen != 0) {
-				bitlen = NBBY - bitlen;
-				if (p1->s6_addr[bytelen] >> bitlen !=
-				    p2->s6_addr[bytelen] >> bitlen)
-				continue;
-			}
-			return ap;
-		}
-	}
-	return NULL;
-}
-
-void
-ipv6rs_handleifa(int cmd, const char *ifname,
-    const struct in6_addr *addr, int flags)
-{
-	struct ra *rap;
-
-	TAILQ_FOREACH(rap, &ipv6_routers, next) {
-		if (strcmp(rap->iface->name, ifname))
-			continue;
-		ipv6_handleifa_addrs(cmd, &rap->addrs, addr, flags);
-	}
-}
-
 void
 ipv6rs_expire(void *arg)
 {
@@ -1126,7 +809,7 @@ ipv6rs_expire(void *arg)
 	struct ra *rap, *ran;
 	struct ra_opt *rao, *raon;
 	struct timeval now, lt, expire, next;
-	int expired, valid;
+	int expired;
 
 	ifp = arg;
 	get_monotonic(&now);
@@ -1139,110 +822,56 @@ ipv6rs_expire(void *arg)
 		lt.tv_sec = rap->lifetime;
 		lt.tv_usec = 0;
 		timeradd(&rap->received, &lt, &expire);
-		if (rap->lifetime == 0 || timercmp(&now, &expire, >)) {
-			valid = 0;
-			if (!rap->expired) {
-				syslog(LOG_WARNING,
-				    "%s: %s: router expired",
-				    ifp->name, rap->sfrom);
-				rap->expired = expired = 1;
-				ipv6ns_cancelproberouter(rap);
-			}
-		} else {
-			valid = 1;
-			timersub(&expire, &now, &lt);
-			if (!timerisset(&next) || timercmp(&next, &lt, >))
-				next = lt;
+		if (timercmp(&now, &expire, >)) {
+			syslog(LOG_INFO, "%s: %s: expired Router Advertisement",
+			    ifp->name, rap->sfrom);
+			rap->expired = expired = 1;
+			continue;
 		}
-
-		/* Addresses are expired in ipv6ns_probeaddrs
-		 * so that DHCPv6 addresses can be removed also. */
-
+		timersub(&expire, &now, &lt);
+		if (!timerisset(&next) || timercmp(&next, &lt, >))
+			next = lt;
+	
 		TAILQ_FOREACH_SAFE(rao, &rap->options, next, raon) {
-			if (rap->expired) {
-				switch(rao->type) {
-				case ND_OPT_RDNSS: /* FALLTHROUGH */
-				case ND_OPT_DNSSL:
-					/* RFC6018 end of section 5.2 states
-					 * that if tha RA has a lifetime of 0
-					 * then we should expire these
-					 * options */
-					TAILQ_REMOVE(&rap->options, rao, next);
-					expired = 1;
-					free(rao->option);
-					free(rao);
-					continue;
-				}
-			}
 			if (!timerisset(&rao->expire))
 				continue;
 			if (timercmp(&now, &rao->expire, >)) {
-				/* Expired prefixes are logged above */
-				if (rao->type != ND_OPT_PREFIX_INFORMATION)
-					syslog(LOG_WARNING,
-					    "%s: %s: expired option %d",
-					    ifp->name, rap->sfrom, rao->type);
+				syslog(LOG_INFO,
+				    "%s: %s: expired option %d",
+				    ifp->name, rap->sfrom, rao->type);
 				TAILQ_REMOVE(&rap->options, rao, next);
 				expired = 1;
 				free(rao->option);
 				free(rao);
 				continue;
 			}
-			valid = 1;
 			timersub(&rao->expire, &now, &lt);
 			if (!timerisset(&next) || timercmp(&next, &lt, >))
 				next = lt;
 		}
-
-		/* No valid lifetimes are left on the RA, so we might
-		 * as well punt it. */
-		if (!valid && TAILQ_FIRST(&rap->addrs) == NULL)
-			ipv6rs_free_ra(rap);
 	}
 
 	if (timerisset(&next))
-		eloop_timeout_add_tv(&next, ipv6rs_expire, ifp);
+		add_timeout_tv(&next, ipv6rs_expire, ifp);
 	if (expired) {
-		ipv6_buildroutes();
-		script_runreason(ifp, "ROUTERADVERT");
+		ipv6_build_routes();
+		run_script_reason(ifp, "ROUTERADVERT");
 	}
 }
 
 int
 ipv6rs_start(struct interface *ifp)
 {
-	struct rs_state *state;
 
-	syslog(LOG_INFO, "%s: soliciting an IPv6 router", ifp->name);
-	if (sock == -1) {
-		if (ipv6rs_open() == -1) {
-			syslog(LOG_ERR, "%s: ipv6rs_open: %m", __func__);
-			return -1;
-		}
-		eloop_event_add(sock, ipv6rs_handledata, NULL);
-	}
-
-	eloop_timeout_delete(NULL, ifp);
-
-	state = RS_STATE(ifp);
-	if (state == NULL) {
-		ifp->if_data[IF_DATA_IPV6RS] = calloc(1, sizeof(*state));
-		state = RS_STATE(ifp);
-		if (state == NULL) {
-			syslog(LOG_ERR, "%s: %m", __func__);
-			return -1;
-		}
-	}
+	delete_timeout(NULL, ifp);
 
 	/* Always make a new probe as the underlying hardware
 	 * address could have changed. */
 	ipv6rs_makeprobe(ifp);
-	if (state->rs == NULL) {
-		syslog(LOG_ERR, "%s: ipv6rs_makeprobe: %m", __func__);
+	if (ifp->rs == NULL)
 		return -1;
-	}
 
-	state->rsprobes = 0;
+	ifp->rsprobes = 0;
 	ipv6rs_sendprobe(ifp);
 	return 0;
 }
@@ -1252,23 +881,12 @@ ipv6rs_drop(struct interface *ifp)
 {
 	struct ra *rap;
 	int expired = 0;
-	TAILQ_HEAD(rahead, ra) rtrs;
 
-	eloop_timeout_delete(NULL, ifp);
-	TAILQ_INIT(&rtrs);
-	TAILQ_FOREACH(rap, &ipv6_routers, next) {
-		if (rap->iface == ifp) {
+	TAILQ_FOREACH(rap, &ipv6_routers, next)
+		if (rap->iface == ifp)
 			rap->expired = expired = 1;
-			TAILQ_REMOVE(&ipv6_routers, rap, next);
-			TAILQ_INSERT_TAIL(&rtrs, rap, next);
-		}
-	}
 	if (expired) {
-		while ((rap = TAILQ_FIRST(&rtrs))) {
-			TAILQ_REMOVE(&rtrs, rap, next);
-			ipv6rs_drop_ra(rap);
-		}
-		ipv6_buildroutes();
-		script_runreason(ifp, "ROUTERADVERT");
+		ipv6_build_routes();
+		run_script_reason(ifp, "ROUTERADVERT");
 	}
 }
