@@ -1,6 +1,6 @@
-/*
+/* 
  * dhcpcd - DHCP client daemon
- * Copyright (c) 2006-2013 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2011 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -52,12 +52,10 @@ struct ra {
 	struct timeval received;
 	unsigned char flags;
 	uint32_t lifetime;
-	uint32_t reachable;
-	uint32_t retrans;
 	uint32_t mtu;
-	struct ipv6_addrhead addrs;
+	TAILQ_HEAD(, ipv6_addr) addrs;
 	TAILQ_HEAD(, ra_opt) options;
-
+	
 	unsigned char *ns;
 	size_t nslen;
 	int nsprobes;
@@ -67,32 +65,15 @@ struct ra {
 
 extern TAILQ_HEAD(rahead, ra) ipv6_routers;
 
-struct rs_state {
-	unsigned char *rs;
-	size_t rslen;
-	int rsprobes;
-};
-
-#define RS_STATE(a) ((struct rs_state *)(ifp)->if_data[IF_DATA_IPV6RS])
-
-#ifdef INET6
+int ipv6rs_open(void);
+void ipv6rs_handledata(void *);
 int ipv6rs_start(struct interface *);
 ssize_t ipv6rs_env(char **, const char *, const struct interface *);
-const struct ipv6_addr * ipv6rs_findprefix(const struct ipv6_addr *);
-int ipv6rs_addrexists(const struct ipv6_addr *);
 void ipv6rs_freedrop_ra(struct ra *, int);
 #define ipv6rs_free_ra(ra) ipv6rs_freedrop_ra((ra),  0)
 #define ipv6rs_drop_ra(ra) ipv6rs_freedrop_ra((ra),  1)
 ssize_t ipv6rs_free(struct interface *);
 void ipv6rs_expire(void *arg);
 int ipv6rs_has_ra(const struct interface *);
-void ipv6rs_handleifa(int, const char *, const struct in6_addr *, int);
 void ipv6rs_drop(struct interface *);
-#else
-#define ipv6rs_start(a) {}
-#define ipv6rs_free(a)
-#define ipv6rs_has_ra(a) 0
-#define ipv6rs_drop(a)
-#endif
-
 #endif
